--[[
    对讲模块( by：zhang.Yi && hyl )
        1. 采用webrtc做数据传输方式，
            主动连接方为客户端模式，被动接收方为服务端模式
		2. 对接web api，
			2.1 mdns 扫描(设备名、ip、port)
			2.2 web和设备对讲测试，即web和设备建立webrtc会话，测试双向语音
			2.3 添加设备与设备间建立webrtc会话，
		3. 音频数据的输入输出
            3.1 采用USB耳机
                定时器查询 USB耳机设备接入和移除

            3.2 NDI设备还可以采用FPGA模拟音频

        4. 业务逻辑
            4.1 web与设备连接 和 设备与设备之间连接 优先级一样
            4.2 定时检查是否需要连接 或者 webrtc连接的状态
--]]

local coaf = require "coaf.core"
local DC = require "supertable.declare"
local CONF = require "supertable.core"
local thread = require "coaf.thread"
local _dbg = require "coaf.debug"
local client = require "coaf.client"

local cjson = require "cjson"
local ltn12 = require("ltn12")

local pl = require "pl.pretty" --打印
local funcs = require "nm_funcs"
local utilBase64 = require "utilBase64"
local httpclient = require "httpclient"
require "kmCore"

--local sys_audio_config = require "sys_audio_config" -- 加载系统音频配置

local _M = {
    talkConfig = nil, --对讲的配置文件，启动的时候就去读取
    devices = {},
    hwParam = {}, --硬件设备的参数

    DeviceMode = "", --本设备的型号

    webRtcTalk = nil,  -- webrtc对讲传输通道
    talkId = 0,  -- 对讲ID
    talkNeedDel = 0, -- 对讲链接需要删除

    audioTalkEng = nil, -- 音频对讲编码 
    swMediaSource = nil, -- 交换媒体源
    usbTalkChannel = nil, -- USB 设备 
    usbTalkCard = nil, -- USB 对讲设备

    audioSink = nil, --音频对讲解码
    audioDecoder = nil,
    audiorSession = nil,

    delayedDiscovery = 0,
    discoverCount = 0,
    idSeed = 0,

    threadId = nil, -- USB设备监听线程
    runing = true,

    deviceStatus = {}, --扫描到设备的所有信息

    audioMode = nil, --音频格式 G711 / OPUS
    headsetInput = nil, --耳机设备输入方式 USB / 3.5MM(FPGA)

    fpgaAnalogTalk = nil,

    connectIp = "", --建立连接的设备IP地址
    connectStatus = ""--建立连接的设备状态
}

local function GetDeviceMode()
    if _M.DeviceMode == "" then
        local _done, features = pcall(require, "oem")
        if _done then
            if type(features) ~= "table" then
                _M.DeviceMode = _G.OEM_TYPE
            else
                _M.DeviceMode = features.OEM_TYPE
            end
        end
        if _M.DeviceMode == "" then
            _done, features = pcall(require, "features")
            if _done then
                if type(features) ~= "table" then
                    _M.DeviceMode = _G.PRODUCT_ID
                else
                    _M.DeviceMode = features.PRODUCT_ID
                end
            end
        end
    end
    return _M.DeviceMode
end

local function GetDeviceName()
    -- local filename = "/etc/hostname"
    -- local f = io.open( filename, "r" )
	-- if f then
	-- 	local r = f:read("*l")
	-- 	f:close()
	-- 	return r
    -- end
    return "KILOVIEW"
end

local function GetHeadsetInput()
    if _M.talkConfig.headset and (_M.talkConfig.headset=="3.5mm" or _M.talkConfig.headset=="3.5MM") then
        return "FPGA"
    end
    return "USB"
end

local function GetHeadsetList()
    -- local list = [{label="USB",value="USB"}, {label="3.5MM",value="3.5MM"}]
    local list = {}
    table.insert(list, {label="USB",value="USB"})
    table.insert(list, {label="3.5MM",value="3.5MM"})
    return list
end

--获取对讲时采用的音频编码格式，默认PCMU
local function GetSoundEncode()
    local _done, features = pcall(require, "features")
    if not _done then
        features = {}
    elseif type(features) ~= "table" then
        features = _G
    end
    for k,v in pairs(features) do
        if k == "FEATURE_INTERCOM_ENCODE" then
            --return v
        end
    end
    return "PCMU"
end
---------------------------------Mdns Discovery--------------------------------------------
local function GetProxy(objPath, service, loc, timeout, ip)
    local proxy
    local cli
    cli = client.New( service, loc, ip, timeout or 3 ) --3 seconds timeout.
    if not cli then
        _dbg.Error( "TALKDISCOVERY", "Fail to create connection to service '" .. (loc or "local") .. "/" .. (service or "nil") )
        return nil
    end
    proxy = cli:CreateProxy(objPath, true) --Exclusive
    if not proxy then
        _dbg.Error( "TALKDISCOVERY", "Fail to create object proxy for '" .. (objPath or "unknown") .. "'" )
        return nil
    end
    return proxy
end

local function DiscoveryAgain()
    local proxy = GetProxy( "/discovery", "_CO_DAEMON.coserver" ) 
    if proxy then
        --_dbg.Info("TALKDISCOVERY", "The device name is:", _M.talkConfig.link_name)
		--#Bug fix: if not set correct link_token, other devices can not connect manual target to this.
		proxy.SetLinkToken(_M.talkConfig.link_token)

        local options = {}
        options.name  = _M.talkConfig.link_name
        options.group = "KILOVIEW"
        options.mode  = GetDeviceMode()
        options.talk  = true
        local r2,done2,err2 = proxy.PublishGlobal(options)
        if not r2 or not done2 then
            _dbg.Error("TALKDISCOVERY", "mDNS publish device failed:", done2, err2)
            proxy:Destroy()
            return true
        end

        --开启discovery
        local r3,err3 = proxy.StartDiscovery()
        if not r3 then
            _dbg.Error("TALKDISCOVERY", "Start mDNS discovery failed:", err3)
            proxy:Destroy()
            return true
        end

        --初始化，将手动的数据告诉coaf
        for k,v in pairs(_M.talkConfig.devices) do
            local r,done,err = proxy.AddManualTarget(tostring(v.id), v.ip, v.port, v.link_token)
            if err and err ~= "exist" then
                _dbg.Error("TALKDISCOVERY", "Add manual device targets failed:", done, err)
                proxy:Destroy()
                return true
            end
        end
        proxy:Destroy()
    else
        _dbg.Error("TALKDISCOVERY", "Create proxy link to coserver:/discovery failed.")
        return true
    end
    return false
end

local function DiscoveryAllDevice()
    local datas={}
    local k,v
    local proxy = GetProxy( "/discovery", "_CO_DAEMON.coserver", "local", 3, "unix:/var/run/coaf.coserver.ctrl" )
    if proxy then
        local r,data,err =  proxy.GetDiscoveryList()
        proxy:Destroy()
        if r and data then
            --名字的话 应该在data.options 里面, 不管设备是否在线，coserver会将我手动添加的信息及自动发现的信息 整合之后告诉我
            -- pl.dump(data)
            -- pl.dump(_M.talkConfig)
            for k,tabs in pairs(data) do
                --只有本组才能被发现
                if tabs.options and tabs.options.group == "KILOVIEW" then 
                    local talkdev={}
                    _M.devices[tabs.id] = tabs.address 
                    -- _dbg.Info("TALKDISCOVERY", "talkdevice id", tabs.id, "address", tabs.address, "port", tabs.port)
                    talkdev.id       = tabs.id
                    talkdev.port     = 80 --默认http 80端口
                    talkdev.ip       = tabs.address --设备ip地址
                    talkdev.mode     = tabs.options.mode or "" --设备类型
                    talkdev.selected = false --是否被选中了
                    talkdev.status   = tabs.status
                    talkdev.manual   = tabs.manual or false --是否为手动添加的设备
                    talkdev.name     = tabs.options.name or ""

                    -- nousb-没有接入USB设备, ready-准备好, trying-没准备好, talking-此设备对讲中, connection-连接建立,  busy-设备正忙
                    if _M.talkConfig.selected_ip ~= "0.0.0.0" and _M.talkConfig.selected_ip == tabs.address then
                        talkdev.selected = true
                        talkdev.status   = _M.connectStatus
                    end

                    table.insert(datas, talkdev)
                end                  
            end
        end
    else
        _dbg.Error("TALKDISCOVERY", "Create proxy link to coserver:/discovery failed.")
        return false, "system ERROR"
    end
    table.sort( datas, function(a,b)
        return a.id < b.id
    end)
    -- 配置本机信息
    local status = {}
    status.id        = "0"
    status.port      = 80 --默认http 80端口
    status.ip        = "192.168.0.1" --设备ip地址
    status.mode      = GetDeviceMode() --设备类型
    status.status    = "ready"
    status.name      = GetDeviceName() --本机设备名
    status.connectIp = _M.connectIp --建立连接的对端IP地址
    status.device    = false
    status.headset   = "USB"
    if _M.talkConfig.headset then
        status.headset = _M.talkConfig.headset
    end
    status.headsetList = GetHeadsetList()
    if _M.usbTalkCard then
        status.device = true
    end
    if _M.connectIp ~= "" then
        status.status = "connection"
    end
    table.insert(datas, 1, status)
    -- print("----------------------------------------------------")
    -- pl.dump(datas)
    _M.deviceStatus = datas;
end

--更新反馈给web的状态
local function UpdateStatus( ip, status )
    for k,tabs in pairs(_M.deviceStatus) do
        if ip == tabs.ip then
            print("UpdateStatus: ", tabs.ip, " status: ", status);
            tabs.status = status
            break;
        end
    end
    --更新本机设备的状态
    if #_M.deviceStatus>=1 and _M.deviceStatus[1].id and _M.deviceStatus[1].id == "0" then
        _M.deviceStatus[1].connectIp = _M.connectIp
        if _M.connectIp ~= "" then
            _M.deviceStatus[1].status = "connection"
        end
    end
end
--更新反馈给web的状态
local function UpdateSelected( ip, enable )
    for k,tabs in pairs(_M.deviceStatus) do
        if ip == tabs.ip then
            print("UpdateSelected: ", tabs.ip, " enable: ", enable);
            tabs.selected = enable
            break;
        end
    end
end
------------------------------USB设备检查-----------------------------------------------
local function file_line( filename )
	local f = io.open( filename, "r" )
	if f then
		local r = f:read("*l")
		f:close()
		return r
	else
		return nil
	end
end
--[[
    检查USB声音设备
--]]
local function detect_usb_sound_device(ignore_device)
	local item
	local card_id, card_name, card_fnd
	local lfs = require "lfs"
	for item in lfs.dir("/sys/class/sound") do
		if item:match("card%d+") then
			card_id = tonumber(file_line("/sys/class/sound/" .. item .. "/number"))
            card_name = file_line("/sys/class/sound/" .. item .. "/id")
            --_dbg.Info( "TALKBACK", " detect: ", item, ". card_id: ",card_id, ". card_name: ",card_name)
			if card_id and card_name and ignore_device and ignore_deivce ~= "" and ignore_device ~= card_name then
				card_name = nil
			end
            --_dbg.Info( "TALKBACK", " detect card_name: ",card_name)
			if card_name then -- Check is USB device?
				if tonumber( "0x" .. (file_line("/sys/class/sound/" .. item .. "/device/bInterfaceClass") or "#")) then
					card_fnd = card_id
					break
				end
			end
		end
	end
	return card_fnd
end

local function check_sound_bus_id(card_id)                                                                                                                                                                                                                     
    local f = io.popen("/usr/bin/readlink -f /sys/class/sound/card" .. card_id, "r" )                                                         
    if f then                                                                                                                                 
        local path = f:read("*l")                                                                                                                                                                                                                                      
        f:close()                                                                                                                                 
        if not path then                                                                                                                          
            return nil, nil                                                                                                                           
        end                                                                                                                                       
        local dev_path = path:match("^(.*/usb[0-9]+/[0-9]+%-[0-9]+)/.*")                                                                          
        if not dev_path then                                                                                                                                                                                                                                           
            return nil, nil                                                                                                                                                                         
        end                                                                                                                                                                            
            return tonumber(file_line(dev_path .. "/busnum")), tonumber(file_line(dev_path .. "/devnum"))                                                                                      
    else                                                                                                                                                                               
        return nil, nil                                                                                                                                                                                                                                                
    end                                                                                                                                                                                                                                                            
end
------------------获取硬件设备的参数------------------------
local function hw_param_load(card_id)
    local file = "/proc/asound/card"..card_id.."/stream0"
    local device = {
        palyback = {},
        capture = {}
    }
    local p = nil
    for line in io.lines(file) do
        if string.find(line, ":") then
            local reps = ':'
            local list = {}
            for word in string.gmatch(line, '[^'..reps..']+') do 
                word = word:match("^[%s]*(.-)[%s]*$")
                table.insert(list, word)
            end
            if #list == 1 then
                p = nil
                if string.lower(list[1]) == "playback" then
                    p = "palyback";
                elseif string.lower(list[1]) == "capture" then
                    p = "capture"
                end
            end
            if p and #list==2 then
                --提取rates 只取其中一个
                local lo = string.lower(list[1])
                if lo == "rates" then
                    if string.find(list[2], "48000") then
                        device[p][lo] = 48000
                    elseif string.find(list[2], "44100") then
                        device[p][lo] = 44100
                    elseif string.find(list[2], "32000") then
                        device[p][lo] = 32000
                    elseif string.find(list[2], "16000") then
                        device[p][lo] = 16000
                    elseif string.find(list[2], "8000") then
                        device[p][lo] = 8000
                    end
                elseif lo == "channels" then
                    device[p][lo] = tonumber(list[2])
                else
                    device[p][lo] = list[2]
                end
            end
        end
    end
    return device
end
local function hw_param_get(use)
    local channels = 2
    if _M.hwParam[use] and _M.hwParam[use]["channels"] then
        channels = _M.hwParam[use]["channels"]
    end
    local sampling = 48000
    if _M.hwParam[use] and _M.hwParam[use]["rates"] then
        sampling = _M.hwParam[use]["rates"]
    end
    return channels, sampling
end
-----------------------------对讲会话---------------------------------
local function talk_channel_create(id)
    --创建音频Sink
    if not _M.audioSink then
        _dbg.Info( "talk_channel_create" , "audioSink")
        if _M.headsetInput == "USB" then
            if _M.usbTalkCard then
                -- USB设备采用ALSA
                local channels, sampling = hw_param_get("palyback")
                print("CreateAudioSink. palyback ", channels, sampling);
                _M.audioSink = KM_STREAMING.AudioEngine.CreateAudioSink( "ALSA", 
                                                                                { device = "hw:" .. _M.usbTalkCard, 
                                                                                fixed_channels = channels, 
                                                                                fixed_sampling = sampling } )
            end
        else
            -- FPGA
            _M.audioSink = KM_STREAMING.AudioEngine.CreateAudioSink( "FPGA", { device = "fpga",
                                                                                fixed_channels = 2,
                                                                                fixed_sampling = 48000,
                                                                                fixed_sample_size = 768 } )
        end
    end

    -- 创建音频解码
    if not _M.audioDecoder then
        _dbg.Info( "talk_channel_create" , "audioDecoder")
        _M.audioDecoder  = KM_STREAMING.Audio.Decoder.createNew( "play" )
    end

    if _M.audioDecoder and _M.audioSink then
        local rtc = _M.webRtcTalk:GetMediaSource(id)
        if rtc then
            _M.audiorSession = _M.audioDecoder:AddSession( rtc, { sink = _M.audioSink, syncIdr = "ID"..id } )
            if _M.audiorSession then
                _M.audioDecoder:StartSession( _M.audiorSession )
                _dbg.Info( "TalkBack" , "StartSession")
                return true
            end
        end
    end
    return false
end
local function talk_channel_destroy()
    --释放音频解码
    if _M.audioDecoder then
        if _M.audiorSession then
            _dbg.Info( "talk_channel_destroy" , "audiorSession")
            _M.audioDecoder:StopSession( _M.audiorSession )
            _M.audioDecoder:RemoveSession( _M.audiorSession )
            _M.audiorSession = nil
        end
        _dbg.Info( "talk_channel_destroy" , "audioDecoder")
        _M.audioDecoder  = nil --智能指针，赋值为nil就是资源回收
    end
    --释放音频Sink
    if _M.audioSink then
        _dbg.Info( "talk_channel_destroy" , "audioSink")
        KM_STREAMING.AudioEngine.FreeAudioSink(_M.audioSink)
        _M.audioSink = nil;
    end
end
-----------------------------HTTP通信--------------------------------
local function toint(n)
    local s = tostring(n)
    local i,j = s:find('%.')
    if i then
        return tonumber(s:sub(1,i-1))
    else
        return n
    end
end

local function http_get(path, body, timeout)
    local r, body = httpclient.post( path, body, timeout, 147258 )
    if r == 0 then
        while true do 
            coaf.Sleep(0.1)
            r, body = httpclient.result( 147258 )
            if r ~= -1 then
                break;
            end
        end
    end
    return r,body
end

local function talk_connect()
    local param = {
        ip = _M.talkConfig.selected_ip
    }
    local cc, tt = http_get( "http://".._M.talkConfig.selected_ip.."/intercom/talkback/talk_add.json", cjson.encode(param), 2000 )
    if cc==0 and tt and tt ~= "" then
        print(tt)
        local dec = cjson.decode(tt)
        if dec["data"] and dec["data"]["sdp"] then
            local id, sdp =  _M.webRtcTalk:AddSessionClient( _M.swMediaSource:GetMediaSource(), {sdp = dec["data"]["sdp"]})
            _dbg.Info( "TalkBack" , "webRtcTalk AddSessionClient: ", id)
            if id and sdp then
                _M.talkId        = id --记录ID号
                _M.connectStatus = "connection"
                _M.connectIp     = _M.talkConfig.selected_ip
                --client端通过http发送answer sdp给server端，
                    --参数sdp指：client的sdp
                    --参数port指：server的port
                local param2 = {
                    sdp = utilBase64.encodeURI(sdp),
                    port = toint(dec.data.id)
                }
                local cc2, tt2 = http_get( "http://".._M.talkConfig.selected_ip.."/intercom/talkback/talk_answersdp.json", cjson.encode(param2), 2000 )
                _dbg.Info( "TalkBack" , "webRtcTalk http cc2 ", cc2, "tt2 ", tt2)
            else
                --本机sdp连接失败
            end
        else
            if dec["status"] then
                _M.connectStatus = "busy"
            end
        end
    end
end

-----------------------------循环线程----------------------------------
local function loop()
    local discovery = true
    while _M.runing do
        if _M.headsetInput == "USB" then
            if not _M.usbTalkCard then --Check the USB sound card
                _M.usbTalkCard = detect_usb_sound_device()
                if _M.usbTalkCard then
                    local busnum, devnum = check_sound_bus_id(_M.usbTalkCard)
                    if busnum and devnum then
                        os.execute("/usr/bin/usbreset /dev/bus/usb/" .. string.format("%03d", busnum) .. "/" .. string.format("%03d", devnum))
                        coaf.Sleep(3);
                        _M.hwParam = hw_param_load(_M.usbTalkCard)
                        pl.dump(_M.hwParam)
                    end
                    local channels, sampling = hw_param_get("capture")
                    print("CreateAudioSource. capture ", channels, sampling);
                    -- Update the USB sound ALSA source.
                    local new_src = KM_STREAMING.AudioEngine.CreateAudioSource( "ALSA", {
                    	device = "hw:" .. _M.usbTalkCard,
                        fixed_channels = channels,
                        fixed_sampling = sampling 
                    })
                    if new_src and _M.usbTalkChannel then
                        _dbg.Info( "TALKBACK", " ChangeSource ALSA : ", _M.usbTalkCard)
                        --切换到ALSA 的音频源
                        _M.usbTalkChannel:ChangeSource( new_src )
                    end
                end
            else
                -- USB 设备移除
                if not tonumber("0x" .. (file_line("/sys/class/sound/card" .. _M.usbTalkCard .. "/device/bInterfaceClass") or "#")) then
                    _M.usbTalkCard = nil
                    if _M.usbTalkChannel then
                        --切换到null音频源
                        local null_src = KM_STREAMING.AudioEngine.CreateAudioSource( "NULL", { device = "NULL",	fixed_channels = 1 })
                        if null_src then
                            _M.usbTalkChannel:ChangeSource( null_src )
                            print("TalkBack: usbTalkChannel:ChangeSource null_src ")
                        end
                    end
                    --销毁音频解码
                    talk_channel_destroy();
                    --清空硬件参数
                    _M.hwParam = {}
                end
            end
        elseif _M.headsetInput == "FPGA" then
            if not _M.fpgaAnalogTalk then
                local new_src = KM_STREAMING.AudioEngine.CreateAudioSource( "FPGA", {
                            device = "fpga",
                            fixed_channels = 2,
                            fixed_sampling = 48000,
                            fixed_sample_size = 768
                        })
                if new_src and _M.usbTalkChannel then
                    _dbg.Info( "TALKBACK", " ChangeSource fpga source: ")
                    --切换到FPGA 的音频源
                    _M.usbTalkChannel:ChangeSource( new_src )
                end
                _M.fpgaAnalogTalk = 1;
            end
        else
            if not _M.headsetInput then
                --销毁音频解码
                talk_channel_destroy();
                
                if _M.usbTalkCard then
                    _M.usbTalkCard = nil
                    _M.hwParam = {}
                end
                
                if _M.fpgaAnalogTalk then
                    _M.fpgaAnalogTalk = nil
                end

                --切换到null音频源
                local null_src = KM_STREAMING.AudioEngine.CreateAudioSource( "NULL", { device = "NULL",	fixed_channels = 1 })
                if null_src and _M.usbTalkChannel then
                    _M.usbTalkChannel:ChangeSource( null_src )
                end

                _M.headsetInput = GetHeadsetInput()
                print("GetHeadsetInput: ", _M.headsetInput)
            end
        end
        --[[
            如果未建立对讲 且 有配置对讲设备的信息 需要创建
            1. 通过http的api获取对方的sdp信息
            2. 创建webrtc client 会话
        --]]
        if _M.talkId == 0 then
            if _M.talkConfig.selected_ip and _M.talkConfig.selected_ip~="0.0.0.0" then
                talk_connect()
                UpdateStatus( _M.connectIp, _M.connectStatus )
            end
        end
        --[[
            检查当前的会话是否被回收
        --]]
        if _M.talkId ~= 0 then
            if _M.talkNeedDel == 1 or _M.webRtcTalk:CheckSession(_M.talkId) <= 0 then
                _dbg.Info( "TalkBack" , "webRtcTalk RemoveSession: ", _M.talkId)
                local ip = _M.connectIp
                --销毁
                _M.webRtcTalk:RemoveSession(_M.talkId)
                talk_channel_destroy()
                _M.talkId        = 0
                _M.talkNeedDel   = 0;
                _M.connectIp     = "" 
                _M.connectStatus = "ready"
                UpdateStatus( ip, "ready" )
            else
                if not _M.audiorSession then
                    talk_channel_create(_M.talkId)
                end
            end
        end
        coaf.Sleep(1)
        --[[
            定时mdns 扫描
        --]]
        if discovery then
            _M.delayedDiscovery = _M.delayedDiscovery - 1
            if _M.delayedDiscovery == 0 then
                discovery = DiscoveryAgain()
                _M.delayedDiscovery = 10 --更新下一次扫描的时间
            end
        end
        
        _M.discoverCount = _M.discoverCount + 1
        if _M.discoverCount >= 6 then
            DiscoveryAllDevice();
            _M.discoverCount = 1
        end
	end
end
    
---------------------------------API---------------------------------------------
--[[
    初始化
        1. 创建webrtc组
        2. 检查系统音频配置文件，
            2.1 采用USB设备，
            2.2 创建音频对讲通道，采用G711音频格式
            2.3 创建USB设备监听线程
--]]
function _M.Init() 
    _dbg.Info( "TALKBACK", "--Init--")

    -- 加载配置文件
    local cfgTalk = require "talk_config"
    _M.talkConfig = cfgTalk.Load()
    for k,v in pairs(_M.talkConfig.devices) do
        local id = tonumber(string.sub(v.id, string.len("manual")+1, -1))
        if tonumber(_M.idSeed) <= id then
            _M.idSeed = id
        end
    end
    pl.dump(_M.talkConfig)

    -- 创建webrtc组
    _M.webRtcTalk = KM_STREAMING.Webrtc.Pusher.createNew("talk01")

    -- 创建交换媒体源
    _M.swMediaSource = KM_STREAMING.SwitchSource.createNew()

    -- 检查系统音频配置文件
    if true then
        _M.audioTalkEng = KM_STREAMING.AudioEngine.Encoder.createNew()
        if not _M.audioTalkEng then
            _dbg.Error( "AUDIO_ENC", "Fail to create audio engine!" )
        else
            -- Since I can't ensure which USB device current is, so initialize a NONE audio source.
            _M.audioMode = GetSoundEncode()
            print("GetSoundEncode: ", _M.audioMode)
            if _M.audioMode == "PCMU" then
                _M.usbTalkChannel = _M.audioTalkEng:CreateEncodingChannel( { type="NONE" }, { type="G711", bufferSize = 512000, channels=1 } )
                if  _M.usbTalkChannel then
                    _M.usbTalkChannel:SetParameters({
                        sampleSize = 320,
                        channels   = 1,
                        sampling   = 8000,
                        g711Law    = "ULAW",
                        bitrate    = 64000
                    })
                end
            elseif _M.audioMode == "OPUS" then
                -- opus 音频编码
                _M.usbTalkChannel = _M.audioTalkEng:CreateEncodingChannel( { type="CROSS" }, { type=_M.audioMode, bufferSize = 512000, channels=1 } )
                if _M.usbTalkChannel then
                    --opus 音频编码参数设置
                    _M.usbTalkChannel:SetParameters({
                        sampleSize = 960,
                        channels   = 2,
                        sampling   = 48000,
                        bitrate    = 48000
                    })
                end
            end
            if _M.usbTalkChannel then
                local null_src = KM_STREAMING.AudioEngine.CreateAudioSource( "NULL", { device = "NULL",	fixed_channels = 1 })
                if null_src then
                    _M.usbTalkChannel:ChangeSource( null_src )
                    _M.usbTalkChannel:Start()
                    _M.swMediaSource:SelectAudioEncoder( _M.usbTalkChannel:GetEncoder() )-- 获取音频编码，挂载到sw媒体源上
                end
            end
            -- 延时两秒钟开启扫描
            _M.delayedDiscovery = 2
            _M.discoverCount    = 3
            -- 创建监听线程
            _M.threadId = thread.New( loop )
            if _M.threadId then
                _M.threadId:Start( )
            end
        end
    end
end

--[[
    对讲通道创建
        remote
        ip 设备的ip地址
        port 设备采用的端口（非必填）
        source 声音来源，默认USB设备（非必填）
      no_usb  表示未接入USB设备
      busy 表示设备忙线
--]]
function _M.TalkBackChannelCreate(remote, ip, port, source)
    local talk = {}

    if _M.talkId ~= 0 then
        _dbg.Error("TALKDISCOVERY", "talking, create talk failed")
        return false, "busy"
    end
    
    local id, sdp =  _M.webRtcTalk:AddSessionServer( _M.swMediaSource:GetMediaSource(), {ip = ip, port = port, videoTransport = "", audioTransport = "sendrecv", audioMode = _M.audioMode})
    _dbg.Info( "TalkBack server " , "webRtcTalk AddSessionServer: ", ip, " Id:", id, " remoteIP:", remote, " audioMode: ", _M.audioMode)
    -- for ln in sdp:gmatch("(.-)\r\n") do
    --     _dbg.Info( "TalkBack server " , ln )
    -- end
    --[[
        音频数据传输的webrtc通道建立成功，需要创建音频解码，输送到USB设备
    --]]
    if id and sdp then
        talk.sdp    = sdp
        talk.id     = id
        talk_channel_create(id)
        _M.talkId    = id --记录ID号
        _M.connectIp = remote
        --更新本机设备的连接ip地址
        if #_M.deviceStatus>=1 and _M.deviceStatus[1].id and _M.deviceStatus[1].id == "0" then
            _M.deviceStatus[1].connectIp = _M.connectIp
            _M.deviceStatus[1].status    = "connection"
        end
    end
    return true, talk;
end 

--[[
    对讲通道销毁
--]]
function _M.TalkBackChannelDestroy(id)
    _dbg.Info( "TalkBack" , "TalkBackChannelDestroy: ", id)
    if  _M.webRtcTalk then
        --需要判断id 是否是 Create时创建的会话ID号
        --或者需要判断id 是否是 配置文件中的对讲设备
        if id and id~=0 and id~="0" then
            if _M.talkId ~= 0 then
                _M.talkNeedDel = 1;
            end
        end
    end
end

--[[
    手动添加设备时，需要存到数据库中去
--]]
function _M.TalkBackDeviceAdd(details)    
    _M.idSeed          = _M.idSeed + 2
    details.id         = "manual"..tostring(_M.idSeed)
    details.manual     = true
    details.status     = "sleep"
    details.name       = details.name or "对讲设备" 
    details.link_token = details.link_token or "123456789" 
    details.port       = details.port or 14000
    -- pl.dump(details)
    --需要校验下，目前自动发现的数据 中是否存在 details的数据
    for uId,devIp in pairs(_M.devices) do
        if devIp == details.ip then
            _dbg.Error("TALKDISCOVERY", "Device is already exist in list")
            return false, "DeviceOnDiscoveryList"
        end
    end
    --校验添加的ip是否是本机
    local intfs = funcs.EnumNetworks()           
    for i = 1, #intfs do
        local linkinfo = funcs.GetLinkDetail(intfs[i])
        if linkinfo.address == details.ip then
            return false, "DeviceOnDiscoveryList"
        end                                 
    end

    local proxy = GetProxy( "/discovery", "_CO_DAEMON.coserver", "local", 3, "unix:/var/run/coaf.coserver.ctrl" )
    if proxy then
        local r,done,err = proxy.AddManualTarget(tostring(details.id), details.ip, details.port, details.link_token)
        proxy:Destroy()
        if not r or not done then
            _dbg.Error("TALKDISCOVERY", "Add manual target for", details.id, "error:", done, err)
            return false, "AddManualTargetFailed"
        end
    else
        _dbg.Error("TALKDISCOVERY", "Create proxy link to coserver:/discovery failed.")
    end
    
    --更新配置文件
    _M.talkConfig['devices'][details.id] = details
    local r, r2 = _M.talkConfig:__TRIGGER(true)
    return true, "ok"
end

--[[ 
    删除 手动添加的哪些设备/平台信息
--]]
function _M.TalkBackDeviceDel(id)
    for k,v in pairs(_M.talkConfig.devices) do
        if tostring(k) == tostring(id) then
            local proxy = GetProxy( "/discovery", "_CO_DAEMON.coserver", "local", 3, "unix:/var/run/coaf.coserver.ctrl" )
            if proxy then
                local r,done,err = proxy.RemoveManualTarget(tostring(id))
                proxy:Destroy()
                if not r or not done then
                    _dbg.Error("TALKDISCOVERY", "Remove manual target of", id, "error:", done, err)
                    return false,"Remove manual target failed"
                end
                
                --删除的设备是当前选中的，则把selected_ip赋值为初始值，关闭对讲
                if v.address == _M.talkConfig.selected_ip then 
                    _dbg.Info("TALKDISCOVERY", "TalkBackDeviceDel: ", id)
                    UpdateSelected(_M.talkConfig.selected_ip, false);
                    --如果需要取消连接的设备正好是已经连接上的设备，需要关闭当前连接
                    if _M.connectIp == _M.talkConfig.selected_ip then
                        _M.connectStatus = "ready"
                        _M.connectIp     = ""
                        UpdateStatus( _M.talkConfig.selected_ip, "ready" )
                        if _M.talkId ~= 0 then
                            _M.talkNeedDel = 1;
                        end
                    end
                    _M.talkConfig.selected_ip = "0.0.0.0"
                end
            else
		        _dbg.Error("TALKDISCOVERY", "Create proxy link to coserver:/discovery failed.")
                return false,"system ERROR"
            end
            _M.talkConfig.devices[k] = nil --把配置文件里对应设备设置为空
            _M.devices[k] = nil
            break
        end
    end

    --更新配置文件
    local r, r2 = _M.talkConfig:__TRIGGER(true)
    return true
end

--[[
    设置对讲,选中某个对讲设备，
    enable = 1 开启对讲，则将把当前选择的ip赋值给selected_ip
    enable = 0 关闭对讲，则将把selected_ip 清空
--]]
function _M.TalkBackDeviceSelected(id, enable)
    local msg = "ok"
    for k,v in pairs(_M.devices) do
        if tostring(k) == tostring(id) then
            _dbg.Info("TALKDISCOVERY", "selected ip: ", _M.talkConfig.selected_ip, " enable: ", enable)
            if enable == 1 then
                if _M.talkConfig.selected_ip == "0.0.0.0" then
                    _M.talkConfig.selected_ip =_M.devices[k]
                    UpdateSelected(_M.talkConfig.selected_ip, true);
                else
                    msg = "DeviceAlreadySelected"
                end
            else 
                UpdateSelected(_M.talkConfig.selected_ip, false);
                --如果需要取消连接的设备正好是已经连接上的设备，需要关闭当前连接
                if _M.connectIp == _M.talkConfig.selected_ip then
                    _M.connectStatus = "ready"
                    _M.connectIp     = ""
                    UpdateStatus( _M.talkConfig.selected_ip, "ready" )
                    if _M.talkId ~= 0 then
                        _M.talkNeedDel = 1;
                    end
                end
                _M.talkConfig.selected_ip = "0.0.0.0"
            end
            break;
        end
    end
    pl.dump(_M.deviceStatus)
    --更新配置文件
    _M.talkConfig:__TRIGGER(true)
    return msg
end

--[[
    修改 手动添加的哪些设备/平台信息
--]]
function _M.TalkBackDeviceModify(id,details)
    _M.TalkBackDeviceDel(id)
    details.id = id
    _M.TalkBackDeviceAdd(details,id)  
    
    return true
end

function _M.GetAllInfos()
    return _M.deviceStatus
end

function _M.Modify_link_token(link_token) 
    _M.cfg_TalkDevice.link_token = link_token
    local proxy = _M.GetProxy( "/discovery", "_CO_DAEMON.coserver", "local", 3, "unix:/var/run/coaf.coserver.ctrl" )
    if proxy then
        local r2 = proxy.SetLinkToken(link_token)
        if not r2 then
            _dbg.Error("TALKDISCOVERY", "Modify link token to:", link_token, "failed!")
            return false
        end
    else
        _dbg.Error("TALKDISCOVERY", "Create proxy link to coserver:/discovery failed.")
        return false, "system ERROR"
    end
    _M.cfg_TalkDevice:__TRIGGER(true)
    return true
end

function _M.Modify_link_name(link_name) 
    _M.cfg_TalkDevice.link_name = link_name
    local proxy = _M.GetProxy( "/discovery", "_CO_DAEMON.coserver", "local", 3, "unix:/var/run/coaf.coserver.ctrl" )
    if proxy then
        local r2,done2,err2 = proxy.PublishGlobal({link_token=_M.cfg_TalkDevice.link_token,name=_M.cfg_TalkDevice.link_name})
        if not r2 or not done2 then
            _dbg.Error("TALKDISCOVERY", "Modify device name to:", link_name, "error:", done2, err2)
            -- return false
        end
    else
        _dbg.Error("TALKDISCOVERY", "Create proxy link to coserver:/discovery failed.")
        -- return false,"discovery proxy ERROR!!!"
    end
    _M.cfg_TalkDevice:__TRIGGER(true)
    return true
end

--[[
    解析客户端发送过来的answer sdp
--]]
function _M.ParseAnswerSdp(args)
    if not args.port and not args.sdp then
        return false,"sdp is nil"
    end

    --local param = cjson.decode( utilBase64.base64decode(args.param) )
    local ret   =  _M.webRtcTalk:ParseWebrtcSdp( {sdp = utilBase64.decodeURI(args.sdp), port = args.port})
    _dbg.Info( "TalkBack server " , "webRtcTalk ParseWebrtcSdp: ", ret)
    if ret < 0 then
        return false,"parse sdp ERROR"
    end
    return true;
end

function _M.SelectHeadsetInput(args)
    if args and (args=="USB" or args=="usb" or args=="3.5mm" or args=="3.5MM") then
        --更新配置文件
        _M.talkConfig.headset = args;
        _M.talkConfig:__TRIGGER(true)

        --耳机输入修改，需要重新配置
        local headset = GetHeadsetInput()
        if headset ~= _M.headsetInput then
            _M.headsetInput = nil;
        end

        --更新发给web的设备状态信息
        _M.deviceStatus[1].headset = _M.talkConfig.headset;
        
        coaf.Sleep(1);
        return true;
    end
    return false;
end

return _M
